#include <extdll.h>
#include <sdk_util.h>
#include <meta_api.h>

#include "plugin.h"

extern cvar_t *pcv_gm_block_on;
edict_t *pSearchEntity = NULL;
TraceResult tr2;

static edict_t * FindEntityInSphere_Post(edict_t *pEdictStartSearchAfter, const float *org, float rad)
{
	// FindEntityInSphere is always called in a while loop, so on last call it will always returns with NULL.
	pSearchEntity = (META_RESULT_STATUS == MRES_OVERRIDE) ? META_RESULT_OVERRIDE_RET(edict_t*) : META_RESULT_ORIG_RET(edict_t*);

	RETURN_META_VALUE(MRES_HANDLED, 0);
}

static void TraceLine_Post(const float *vecSrc, const float *vecDest, int fNoMonsters, edict_t *pentToSkip, TraceResult *tr)
{
	if (FNullEnt(pSearchEntity))
		RETURN_META(MRES_IGNORED);

	META_RES ret = MRES_IGNORED;

	if (tr->fStartSolid && pcv_gm_block_on->value > 0.0)
	{
#ifdef _DEBUG
		LOG_MESSAGE(PLID, "Hit Id: %d, Search Id: %d, Fraction: %f, StartSolid: %d, AllSolid: %d", ENTINDEX(tr->pHit), ENTINDEX(pSearchEntity), tr->flFraction, tr->fStartSolid, tr->fAllSolid);
		const char* logformat = "Ghost mine explosion detected and blocked. (bug type: %i, %s bug)";
#endif

		// Bug number 2: TraceLine starts from solid (BSP object) and doesn't hits anything on it's way.
		// fStartSolid is set to '1', fAllSolid is set to '1' and flFraction is set to '1.0'.
		// More over, pHit is set to 'worldspawn' and vecEndPos is correctly set to spot to where we asked it to trace to.
		// Bug here may be in:
		// that engine sets fAllSolid to '1' (part of the path is not in the wall) or
		// that we doesn't hit an entity or
		// that engine doesn't set flFraction and vecEndPos correctly when it hits same BSP object.
		if (tr->pHit != pSearchEntity && tr->flFraction == 1.0)
		{
#ifdef _DEBUG
			LOG_MESSAGE(PLID, logformat, 2, "engine");
#endif
			// Bypass checks in game dll
			tr->flFraction = 0.1f;
			ret = MRES_HANDLED;
		}

		// Bug number 1: Explosion has been happen inside BSP object.
		// But trace results are the same as for explosion inside a player, so we just trace back and see what we will hit.
		else if (tr->pHit == pSearchEntity || tr->flFraction == 1.0)
		{
			(*g_engfuncs.pfnTraceLine)(tr->vecEndPos, vecSrc, fNoMonsters, pentToSkip, &tr2);

			if (tr2.flFraction != 1.0 && !tr2.fStartSolid && tr2.pHit != pSearchEntity)
			{
#ifdef _DEBUG
				LOG_MESSAGE(PLID, logformat, 1, "gamedll");
#endif
				// Bypass checks in game dll
				tr->flFraction = 0.1f;
				tr->pHit = tr2.pHit;
				ret = MRES_HANDLED;
			}
		}
	}

	// Only need fix for the very first "TraceLine" after "FindEntityIsSphere"
	pSearchEntity = NULL;
	RETURN_META(ret);
}

int plugin_GetEngineFunctions_Post(enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
	memset(pengfuncsFromEngine, 0, sizeof(enginefuncs_t));

	pengfuncsFromEngine->pfnFindEntityInSphere = FindEntityInSphere_Post;
	pengfuncsFromEngine->pfnTraceLine = TraceLine_Post;

	return(TRUE);
}
